/*
 * Paraz - a simple tool for harvesting performance data for  parallel
 * operations using Z39.50.
 * Copyright (c) 2000 Index Data Aps
 * See LICENSE file for details.
 */

#define PARAZ_VERSION "0.0"
#define MAX_DATABASES 100

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/socket.h>

#include <yaz/comstack.h>
#include <yaz/tcpip.h>
#include <yaz/proto.h>
#include <yaz/readconf.h>
#include <yaz/pquery.h>

#include "eventl.h"

static char *myname;
static long int runId=-1;
static int n_conn = 0; /* number of open connections */
static int n_targ = 0; /* number of targets in the test */

struct command
{
    enum           /* what to do */
    {
	Connect,	/* Connecting */
	Init,
	Search,
	Present,
	Disconnect,
	Mark,		/* Set a mark in the output file */
	Wait		/* Waiting (for other targets to complete) */
    } what;
    union      	/* how to do it */
    {
	struct search
	{
	    Z_Query *query;
	} search;
	struct present
	{
	    int howmany;
	} present;
	char mark[128];
	struct waitc
	{
	    int reset;
	} waitc;
    } how;
    struct command *next;
};

static struct command *command_list = 0;

struct target
{
    char fullname[256];
    char hostport[128];
    char *ibuf;
    int ibufsize;
    char databases[MAX_DATABASES][128];
    struct command *current_command;
    COMSTACK link;
    ODR odr_in, odr_out;
    struct target *next;
    void *addr;
    int hits;
    enum target_state
    {
	No_connection,
	Idle,
	Pending,
	Waiting,
	Finished,
	Failed
    } state;
};

static struct target *target_list = 0;
static IOCHAN channel_list = 0;
static ODR odr_global = 0;

static struct parameters {
    int timeout;		/* operations timeout, in seconds */
    char implementationId[128];
    char implementationName[128];
    char implementationVersion[128];
    struct timeval base_time;
} global_parameters = 
{
    30,
    "81",
    "Parallel searching performance tester",
    PARAZ_VERSION,
    {0,0}
};

static void startmark(char *event, char *targ)
{
    struct timeval now;
    long sec=0, usec=0;

    if ( global_parameters.base_time.tv_sec != 0 )
    {
        gettimeofday(&now, 0);
        sec = now.tv_sec - global_parameters.base_time.tv_sec;
        usec = now.tv_usec - global_parameters.base_time.tv_usec;
        if (usec < 0)
        {
        sec--;
        usec += 1000000;
        }
    }
    printf("%10ld,%-10s, %3ld.%6.6ld, %s ",
           runId, event, sec, usec, targ);
}

static void mark2(char *event, char *targ, char *msg, int num1, int num2)
{
    startmark(event,targ);
    if (msg && ! *msg)
      msg=0;
    printf(", %s, %d, %d\n", 
	   msg  ? msg  : "",
	   num1, num2);
} // mark2

static void mark1(char *event, char *targ, char *msg, int num)
{
    startmark(event,targ);
    printf(", %s, %d \n",
       msg, num);
} // mark


static void mark(char *event, char *targ, char *msg)
{
    startmark(event,targ);
    printf(", %s \n", msg);
} // mark

static void mark0(char *event, char *targ)
{
    startmark(event,targ);
    printf("\n");
} // mark


static void headermarks()
{
    time_t tim;
    struct tm *tmp;
    char tbuf[50];

    time(&tim);
    strcpy(tbuf,ctime(&tim));
    tbuf[19]='\0';
          
    tmp = localtime(&tim);
    mark2("DATE","(all)", tbuf+4,
           (tmp->tm_year+1900)*10000+(tmp->tm_mon+1)*100+tmp->tm_mday, /* yyyymmdd */
           tmp->tm_hour*10000+tmp->tm_min*100+tmp->tm_sec); /* hhmmss */
}

static void completemark()
{
    time_t tim;
    char tbuf[50];

    time(&tim);
    strcpy(tbuf,ctime(&tim));
    tbuf[19]='\0';
          
    mark1("COMPLETE","(all)",tbuf+4,n_targ);

}

static int send_apdu(struct target *t, Z_APDU *a)
{
    char *buf;
    int len;

    if (!z_APDU(t->odr_out, &a, 0, 0))
    {
        odr_perror(t->odr_out, "Encoding APDU");
	abort();
    }
    buf = odr_getbuf(t->odr_out, &len, 0);
    if (cs_put(t->link, buf, len) < 0)
    {
        fprintf(stderr, "cs_put: %s", cs_errmsg(cs_errno(t->link)));
        exit(1);
    }
    odr_reset(t->odr_out); /* release the APDU structure  */
    return 1;
}

static void fatal(char *string, char *add)
{
    fprintf(stderr, "%s: %s (%s)\n", myname, string, add ? add : "");
    exit(0);
}

static void read_commands(char *filename)
{
    FILE *inf;
    int lineno, argc;
    char line[512];
    char *argv[32];
    struct target **target_p = &target_list, *target;
    struct command **command_p = &command_list, *command;

    if (!strcmp(filename, "-"))
	inf = stdin;
    else
	fatal("Can only read from stdin (filename '-')", 0);

    while ((argc = readconf_line(inf, &lineno, line, 512, argv, 32)))
    {
	if (!strcmp(argv[0], "target"))
	{
	    char *p;
	    if (argc != 2)
		fatal("target command requires exactly one parameter", 0);
	    n_targ++;
	    *target_p = target = xmalloc(sizeof(**target_p));
	    target->next = 0;
	    target_p = &target->next;
	    target->current_command = 0;
	    target->state = No_connection;
	    target->ibuf = 0;
	    target->ibufsize = 0;
	    target->odr_in = odr_createmem(ODR_DECODE);
	    target->odr_out = odr_createmem(ODR_ENCODE);
	    target->hits = 0;
	    strcpy(target->fullname, argv[1]);
	    if ((p = strchr(argv[1], '/')))
	    {

		*p = '\0';
		strcpy(target->hostport, argv[1]);
		p++;
		strcpy(target->databases[0], p);
		target->databases[1][0] = '\0';
	    }
	    else
	    {
		strcpy(target->hostport, argv[1]);
		strcpy(target->databases[0], "Default");
		target->databases[1][0] = '\0';
	    }
	}
	else if (!strcmp(argv[0], "mark"))
	{
	    *command_p = command = xmalloc(sizeof(*command));
	    command->next = 0;
	    command_p = &command->next;

	    command->what = Mark;
	    if (argc > 1)
		strcpy(command->how.mark, argv[1]);
	    else
		command->how.mark[0] = '\0';
	}
	else if (!strcmp(argv[0], "connect"))
	{
	    *command_p = command = xmalloc(sizeof(*command));
	    command->next = 0;
	    command_p = &command->next;

	    command->what = Connect;
	}
	else if (!strcmp(argv[0], "init"))
	{
	    *command_p = command = xmalloc(sizeof(*command));
	    command->next = 0;
	    command_p = &command->next;

	    command->what = Init;
	}
	else if (!strcmp(argv[0], "search"))
	{
	    char qbuf[512];
	    int i;
	    Z_RPNQuery *rpnquery;

	    *command_p = command = xmalloc(sizeof(*command));
	    command->next = 0;
	    command_p = &command->next;

	    command->what = Search;
	    qbuf[0] = '\0';
	    for (i = 1; i < argc; i++)
	    {
		strcat(qbuf, argv[i]);
		strcat(qbuf, " ");
	    }
	    
	    command->how.search.query = odr_malloc(odr_global, sizeof(Z_Query));
	    command->how.search.query->which = Z_Query_type_1;
	    command->how.search.query->u.type_1 = rpnquery =
		p_query_rpn (odr_global, PROTO_Z3950, qbuf);
	    if (!rpnquery)
	    {
		fatal("Prefix query error", qbuf);
	    }
	}
	else if (!strcmp(argv[0], "present"))
	{
	    *command_p = command = xmalloc(sizeof(*command));
	    command->next = 0;
	    command_p = &command->next;
	    if (argc < 2)
		command->how.present.howmany = 10;
	    else
                command->how.present.howmany = atoi(argv[1]);

	    command->what = Present;
	}
	else if (!strcmp(argv[0], "wait"))
	{
	    *command_p = command = xmalloc(sizeof(*command));
	    command->next = 0;
	    command_p = &command->next;

	    if (argc == 2 && !strcmp(argv[1], "reset"))
		command->how.waitc.reset = 1;
	    else
		command->how.waitc.reset = 0;

	    command->what = Wait;
	}
	else if (!strcmp(argv[0],"runid"))
	{
	    runId = atol(argv[1]);
	}
	else
	    fatal("Unknown command in input", argv[0]);
    }
    if (!command_list || command_list->what != Connect)
	fatal("First action should be connect", 0);
    if (inf != stdin)
	fclose(inf);
}

static void send_init(IOCHAN i)
{
    struct target *t = iochan_getdata(i);
    Z_APDU *a = zget_APDU(t->odr_out, Z_APDU_initRequest);

    a->u.initRequest->implementationId = global_parameters.implementationId;
    a->u.initRequest->implementationName = global_parameters.implementationName;
    a->u.initRequest->implementationVersion =
	global_parameters.implementationVersion;
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_search);
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_present);
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_namedResultSets);
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_triggerResourceCtrl);
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_scan);
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_sort);
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_extendedServices);
    ODR_MASK_SET(a->u.initRequest->options, Z_Options_delSet);

    ODR_MASK_SET(a->u.initRequest->protocolVersion, Z_ProtocolVersion_1);
    ODR_MASK_SET(a->u.initRequest->protocolVersion, Z_ProtocolVersion_2);
    ODR_MASK_SET(a->u.initRequest->protocolVersion, Z_ProtocolVersion_3);
    if (send_apdu(t, a))
    {
	iochan_setflags(i, EVENT_INPUT);
	t->state = Pending;
    }
    else
    {
	iochan_destroy(i);
	t->state = Failed;
	cs_close(t->link);
    }
}

static void send_search(IOCHAN i)
{
    struct target *t = iochan_getdata(i);
    Z_APDU *a = zget_APDU(t->odr_out, Z_APDU_searchRequest);
    int ndb;
    char **databaselist;

    a->u.searchRequest->query = t->current_command->how.search.query;
    for (ndb = 0; *t->databases[ndb]; ndb++)
	;
    databaselist = odr_malloc(t->odr_out, sizeof(char*) * ndb);
    for (ndb = 0; *t->databases[ndb]; ndb++)
	databaselist[ndb] = t->databases[ndb];

    a->u.searchRequest->databaseNames = databaselist;
    a->u.searchRequest->num_databaseNames = ndb;
    if (send_apdu(t, a))
    {
	iochan_setflags(i, EVENT_INPUT);
	t->state = Pending;
    }
    else
    {
	iochan_destroy(i);
	t->state = Failed;
	cs_close(t->link);
    }
    odr_reset(t->odr_out);
}

static void send_present(IOCHAN i)
{
    struct target *t = iochan_getdata(i);
    Z_APDU *a = zget_APDU(t->odr_out, Z_APDU_presentRequest);
    int toget;

    toget = t->current_command->how.present.howmany;
    if (toget > t->hits)
	toget = t->hits;
    if (!toget)
    {
	t->current_command = t->current_command->next;
	t->state = Idle;
	iochan_setflags(i, EVENT_OUTPUT);
	return;
    }
    a->u.presentRequest->numberOfRecordsRequested = &toget;
    if (send_apdu(t, a))
    {
	iochan_setflags(i, EVENT_INPUT);
	t->state = Pending;
    }
    else
    {
	iochan_destroy(i);
	t->state = Failed;
	cs_close(t->link);
    }
    odr_reset(t->odr_out);
}

static void do_mark(IOCHAN i)
{
    struct target *t = iochan_getdata(i);

    mark("MARK", t->fullname, t->current_command->how.mark);
    t->current_command = t->current_command->next;
    t->state = Idle;
    iochan_setflags(i, EVENT_OUTPUT);
    return;
}

static void do_wait(IOCHAN i)
{
    struct target *t = iochan_getdata(i);

    t->state = Waiting;
    iochan_setflags(i, 0);
}

static void do_initResponse(IOCHAN i, Z_APDU *a)
{
    struct target *t = iochan_getdata(i);
    Z_InitResponse *r = a->u.initResponse;

    if (*r->result)
    {
	mark0("INIT", t->fullname);
	t->state = Idle;
	t->current_command = t->current_command->next;
    }
    else
    {
	mark(/*"FAILED"*/ "INITFAIL", t->fullname, "init-rejected");
	t->state = Failed;
	iochan_destroy(i);
	cs_close(t->link);
    }
}

#if 0
static char *search_geterror(Z_SearchRequest *r)
{
#endif
    

static void do_searchResponse(IOCHAN i, Z_APDU *a)
{
    struct target *t = iochan_getdata(i);
    Z_SearchResponse *r = a->u.searchResponse;

    if (*r->searchStatus)
    {
	mark1("SEARCH", t->fullname, "", *r->resultCount);
	t->hits = *r->resultCount;
    }
    else
    {          /*"FAILED"*/
	mark ("SEARCHFAIL", t->fullname, /*search_geterror(r)*/ "search error");
	t->hits = 0;
    }
    t->state = Idle;
    t->current_command = t->current_command->next;
}

static void do_presentResponse(IOCHAN i, Z_APDU *a)
{
    struct target *t = iochan_getdata(i);
    Z_PresentResponse *r = a->u.presentResponse;

    if (!*r->presentStatus)
	mark1("PRESENT", t->fullname, "",*r->numberOfRecordsReturned);
    else
	mark("PRESFAIL", t->fullname, "failed");
    t->state = Idle;
    t->current_command = t->current_command->next;
}

static void handler(IOCHAN i, int event)
{
    struct target *t = iochan_getdata(i);
    static int waiting = 0;

    if (t->state == No_connection) /* Start connection */
    {
	int res = cs_connect(t->link, t->addr);

	t->state = Pending;
	if (!res) /* we are go */
	    iochan_setevent(i, EVENT_OUTPUT);
	else if (res == 1)
	    iochan_setflags(i, EVENT_OUTPUT);
	else
	{
	    //printf("ERROR %s connect\n", t->hostport);
	    mark0("CONNFAIL",t->hostport);
	    cs_close(t->link);
	    t->state = Failed;
	    iochan_destroy(i);
	    return;
	}
    }
    else if (t->current_command && t->current_command->what == Connect &&
	event & EVENT_OUTPUT)
    {
	int errcode, errlen = sizeof(errcode);

	if (getsockopt(cs_fileno(t->link), SOL_SOCKET, SO_ERROR, &errcode,
	    &errlen) < 0 || errcode != 0)
	{
	    mark(/*"FAILED"*/ "CONNFAIL", t->fullname, "connect");
	    cs_close(t->link);
	    iochan_destroy(i);
	    t->state = Failed;
	    return;
	}
	else
	{
	    mark0("CONNECTED", t->fullname);
	    t->current_command = t->current_command->next;
	    t->state = Idle;
	}
    }
    else if (event & EVENT_INPUT)
    {
	int len = cs_get(t->link, &t->ibuf, &t->ibufsize);

	if (len < 0)
	{
	    mark("LINKFAIL" /*"FAILED"*/, t->fullname, "read-error");
	    cs_close(t->link);
	    iochan_destroy(i);
	    t->state = Failed;
	    return;
	}
	if (len == 0)
	{
	    mark("LINKFAIL", t->fullname, "connection-closed");
	    cs_close(t->link);
	    iochan_destroy(i);
	    t->state = Failed;
	    return;
	}
	else if (len > 1)
	{
	    Z_APDU *a;

	    odr_reset(t->odr_in);
	    odr_setbuf(t->odr_in, t->ibuf, len, 0);
	    if (!z_APDU(t->odr_in, &a, 0, 0))
	    {
		mark("ERROR", t->fullname, "protocol");
		cs_close(t->link);
		iochan_destroy(i);
		t->state = Failed;
		return;
	    }
	    switch (a->which)
	    {
		case Z_APDU_initResponse:
		    do_initResponse(i, a);
		    break;
		case Z_APDU_searchResponse:
		    do_searchResponse(i, a);
		    break;
		case Z_APDU_presentResponse:
		    do_presentResponse(i, a);
		    break;
		default:
		    mark("ERROR", t->fullname, "unexpected-package");
		    cs_close(t->link);
		    iochan_destroy(i);
		    t->state = Failed;
		    return;
	     }
	    if (cs_more(t->link))
		iochan_setevent(i, EVENT_INPUT);
	}
	/* if len==1 we do nothing but wait for more input */
    }
    if (!t->current_command)
    {
	cs_close(t->link);
	iochan_destroy(i);
	t->state = Finished;
	n_conn--;
	mark2("FINISHED", t->fullname, "", n_targ, n_targ-n_conn);
    }
    else if (t->state == Idle)
    {
	switch (t->current_command->what)
	{
	    case Init:
	    	send_init(i);
		break;
	    case Mark:
	    	do_mark(i);
		break;
	    case Search:
	    	send_search(i);
		break;
	    case Present:
	    	send_present(i);
		break;
	    case Wait:
		do_wait(i);
		waiting = 1;
		break;
	    default:
	    	abort();
	}
    }
    if (waiting)
    {
	IOCHAN ii;
	for (ii = channel_list; ii; ii = iochan_getnext(ii))
	{
	    struct target *tt = iochan_getdata(ii);
	    if (tt->state != Waiting)
		break;
	}

	if (!ii)
	{
	    int reset = t->current_command->how.waitc.reset;

	    for (ii = channel_list; ii; ii = iochan_getnext(ii))
	    {
		struct target *tt = iochan_getdata(ii);
		tt->state = Idle;
		iochan_setflags(ii, EVENT_OUTPUT);
		tt->current_command = tt->current_command->next;
	    }
	    if (reset)
		gettimeofday(&global_parameters.base_time, 0);
	    mark0("WAIT", "(all)");
	}
    }
}

static void initialize_connections(void)
{
    IOCHAN new;
    struct target *target;

    for (target = target_list; target; target = target->next)
    {
	if (!(target->link = cs_create(tcpip_type, 0, PROTO_Z3950)))
	    fatal("Failed to create comstack", 0);
	if (!(target->addr = cs_straddr(target->link, target->hostport)))
	{
	    printf("ERROR %s bad-address", target->hostport);
	    target->state = Failed;
	    continue;
	}
	new = iochan_create(cs_fileno(target->link), handler, 0);
	iochan_setdata(new, target);
	iochan_setevent(new, EVENT_EXCEPT);
	target->current_command = command_list;
	new->next = channel_list;
	channel_list = new;
	n_conn++;
    }
}

void set_runid()
{
    if (-1 == runId)
        runId=time(0);
} // set_runid

int main(int argc, char **argv)
{
    myname = argv[0];
    odr_global = odr_createmem(ODR_ENCODE); /* just for shared buffers */

    read_commands("-");
    set_runid();
    initialize_connections();
    headermarks();
    gettimeofday(&global_parameters.base_time, 0);
    event_loop(&channel_list);
    completemark();
    sleep(2); /* make sure next run id won't be on the same second */
    exit(0);
}
